"""
Created on Dec 18, 2014

@author: tcanham

"""
import copy

from .type_base import BaseType, ValueType
from .type_exceptions import NotInitializedException, TypeMismatchException


class SerializableType(ValueType):
    """
    Representation of the Serializable type (comparable to the ANY type)

    The serializable type is a container for other instances of
    BaseType, including itself.

    @param param: typename = "SomeTypeName" string
    To preserve member order, the member argument is a list of members and their types:
    @param param: mem_list = [ ("member",<ref to BaseType>, format string, description), ... ]
    OR mem_list = [ ("member",<ref to BaseType>, format string), ... ].
    The member descriptions can be None
    """

    def __init__(self, typename, mem_list=None):
        """
        Constructor
        """
        super().__init__()
        if not isinstance(typename, str):
            raise TypeMismatchException(str, type(typename))
        self.__typename = typename

        new_mem_list = []
        # If the member list is defined, stamp in None for any missing descriptions
        if mem_list:
            new_mem_list = [
                entry if len(entry) == 4 else (entry[0], entry[1], entry[2], None)
                for entry in mem_list
            ]
        # Set the member list then set the value
        self.mem_list = new_mem_list

    def validate(self, val=None):
        """ Validate this object including member list and values """
        # Blank member list does not validate
        if not self.mem_list:
            return
        elif not isinstance(self.mem_list, list):
            raise TypeMismatchException(list, self.mem_list)
        for member_name, member_val, format_string, description in self.mem_list:
            # Check each of these members for correct types
            if not isinstance(member_name, str):
                raise TypeMismatchException(str, type(member_name))
            elif not isinstance(member_val, BaseType):
                raise TypeMismatchException(BaseType, type(member_val))
            elif not isinstance(format_string, str):
                raise TypeMismatchException(str, type(format_string))
            elif description is not None and not isinstance(description, str):
                raise TypeMismatchException(str, type(description))
        # When a value is set and is not empty we need to set the member properties
        if not val:
            return
        # If a value is supplied then check each value against the member list
        for val_member, list_entry in zip(val, self.mem_list):
            _, member_list_val, _, _ = list_entry
            member_list_val.validate(
                val_member
            )  # Insure that the the val_member is consistent with the existing member

    @property
    def mem_list(self):
        """ Gets the member list"""
        return self.__mem_list

    @mem_list.setter
    def mem_list(self, mem_list):
        """
        Sets the member list and validates against the current value.
        """
        self.__mem_list = mem_list
        if mem_list is None:
            self.validate(self.mem_list)

    def serialize(self):
        """ Serializes the members of the serializable """
        if self.mem_list is None:
            raise NotInitializedException(type(self))
        return b"".join(
            [member_val.serialize() for _, member_val, _, _ in self.mem_list]
        )

    def deserialize(self, data, offset):
        """ Deserialize the values of each of the members """
        new_member_list = []
        for entry1, member_val, entry3, entry4 in self.mem_list:
            cloned = copy.copy(member_val)
            cloned.deserialize(data, offset)
            new_member_list.append((entry1, cloned, entry3, entry4))
            offset += member_val.getSize()
        self.mem_list = new_member_list

    @property
    def val(self) -> dict:
        """
        The .val property typically returns the python-native type. This the python native type closes to a serializable
        without generating full classes would be a dictionary (anonymous object). This returns such an object.

        :return dictionary of member names to python values of member keys
        """
        return {
            member_name: member_val.val
            for member_name, member_val, _, _ in self.mem_list
        }

    @val.setter
    def val(self, val: dict):
        """
        The .val property typically returns the python-native type. This the python native type closes to a serializable
        without generating full classes would be a dictionary (anonymous object). This takes such an object and sets the
        member val list from it.

        :param val: dictionary containing python types to key names. This
        """
        values_list = [val[name] for name, _, _, _ in self.mem_list]
        # Member list is the explicit store for storing these values
        for val_member, list_entry in zip(values_list, self.mem_list):
            _, member_list_val, _, _ = list_entry
            member_list_val.val = val_member

    def getSize(self):
        """ The size of a struct is the size of all the members """
        return sum([mem_type.getSize() for _, mem_type, _, _ in self.mem_list])

    def to_jsonable(self):
        """
        JSONable type for a serializable
        """
        members = {}
        for member_name, member_value, member_format, member_desc in self.mem_list:
            members[member_name] = {"format": member_format, "description": member_desc}
            members[member_name].update(member_value.to_jsonable())
        return members



"""
Created on Dec 18, 2014

@author: tcanham

"""

# Exception classes for all types
from fprime.common.error import FprimeException


class TypeException(FprimeException):
    """ An exception in our python types"""

    def __init__(self, val):
        super().__init__(val)
        self.except_msg = val

    def getMsg(self):
        """ Gets the exception message"""
        return self.except_msg


class AbstractMethodException(TypeException):
    """ Not implemented exception under another name """

    def __init__(self, val):
        super().__init__("%s must be implemented since it is abstract!" % str(val))


class TypeRangeException(TypeException):
    """ Value is out of range """

    def __init__(self, val):
        super().__init__("Value %s out of range!" % str(val))


class StringSizeException(TypeException):
    """ String size is to large for defined type """

    def __init__(self, size, max_size):
        super().__init__(
            "String size {} is greater than {}!".format(str(size), str(max_size))
        )


class TypeMismatchException(TypeException):
    """ Wrong type found exception """

    def __init__(self, expected_type, actual_type):
        super().__init__(
            "Type {} expected, type {} found!".format(expected_type, actual_type)
        )


class ArrayLengthException(TypeException):
    """ Array length mismatched """

    def __init__(self, arr_type, expected_len, actual_len):
        super().__init__(
            "Array type %s is of length %s, actual length %s found!"
            % (arr_type, expected_len, actual_len)
        )


class EnumMismatchException(TypeException):
    """ Enum member not defined """

    def __init__(self, enum, bad_member):
        super().__init__(
            "Invalid enum member {} set in {} enum!".format(bad_member, enum)
        )


class DeserializeException(TypeException):
    """ Exception during deserialization """

    pass


class ArgNotFoundException(TypeException):
    """ Argument not found exception """

    def __init__(self, message):
        super().__init__("Arg %s not found!" % message)


class NotInitializedException(TypeException):
    """ Did not initialize types """

    def __init__(self, message):
        super().__init__("Instance %s not initialized!" % message)


class NotOverriddenException(TypeException):
    """ Not implemented exception by another name """

    def __init__(self, message):
        super().__init__(
            "Required base class method not overwritten in type %s!" % message
        )


class ArgLengthMismatchException(TypeException):
    """ Mismatch in args lengths vs definition of args """

    def __init__(self, arg_length_actual, arg_length_given):
        super().__init__(
            "%d args provided, but command expects %d args!"
            % (arg_length_given, arg_length_actual)
        )


class CompoundTypeLengthMismatchException(TypeException):
    """ Compound type fields mismatch """

    def __init__(self, field_length_actual, field_length_given):
        super().__init__(
            "%d fields provided, but compound type expects %d fields!"
            % (field_length_given, field_length_actual)
        )



"""
Tests the serialization and deserialization of all the types in Fw/Python/src/fprime/common/models/serialize/

Created on Jun 25, 2020
@author: hpaulson, mstarch
"""
import json
from collections.abc import Iterable

import pytest

from fprime.common.models.serialize.array_type import ArrayType
from fprime.common.models.serialize.bool_type import BoolType
from fprime.common.models.serialize.enum_type import EnumType
from fprime.common.models.serialize.numerical_types import (
    F32Type,
    F64Type,
    I8Type,
    I16Type,
    I32Type,
    I64Type,
    U8Type,
    U16Type,
    U32Type,
    U64Type,
)
from fprime.common.models.serialize.serializable_type import SerializableType
from fprime.common.models.serialize.string_type import StringType
from fprime.common.models.serialize.time_type import TimeBase, TimeType
from fprime.common.models.serialize.type_base import BaseType, ValueType

from fprime.common.models.serialize.type_exceptions import (
    AbstractMethodException,
    DeserializeException,
    NotInitializedException,
    TypeMismatchException,
    TypeRangeException,
)


PYTHON_TESTABLE_TYPES = [
    True,
    False,
    -1,
    0,
    300,
    "abc",
    "True",
    "False",
    3.1412,
    (0, 1),
    (True, False),
    [],
    [0],
    {},
    {"abc": 123},
    {2, 4, 3},
]


def valid_values_test(type_input, valid_values, sizes, extras=None):
    """ Tests to be run on all types """
    if not isinstance(sizes, Iterable):
        sizes = [sizes] * len(valid_values)
    # Should be able to instantiate a blank type, but not serialize it until a value has been supplied
    if not extras:
        extras = [[]] * len(valid_values)
    instantiation = type_input(*extras[0])
    with pytest.raises(NotInitializedException):
        instantiation.serialize()
    # Should be able to get a JSONable object that is dumpable to a JSON string
    jsonable = instantiation.to_jsonable()
    json.loads(json.dumps(jsonable))

    # Run on valid values
    for value, size, extra in zip(valid_values, sizes, extras):
        instantiation = type_input(*extra, val=value)
        assert instantiation.val == value
        assert instantiation.getSize() == size

        # Check assignment by value
        by_value = type_input(*extra)
        by_value.val = value
        assert by_value.val == instantiation.val, "Assignment by value has failed"
        assert by_value.getSize() == size

        # Check serialization and deserialization
        serialized = instantiation.serialize()
        for offset in [0, 10, 50]:
            deserializer = type_input(*extra)
            deserializer.deserialize((b" " * offset) + serialized, offset)
            assert instantiation.val == deserializer.val, "Deserialization has failed"
            assert deserializer.getSize() == size


def invalid_values_test(
    type_input, invalid_values, exception_class=TypeMismatchException
):
    """ Check invalid values for all types """
    for item in invalid_values:
        # Constructor initialization
        with pytest.raises(exception_class):
            instantiation = type_input(item)
        # Value initialization
        with pytest.raises(exception_class):
            instantiation = type_input()
            instantiation.val = item
        # Deserialization problems required space
        for offset in [0, 10, 50]:
            with pytest.raises(DeserializeException):
                instantiation = type_input()
                instantiation.deserialize(b" " * offset, offset)


def ser_deser_time_test(t_base, t_context, secs, usecs):
    """
    Test serialization/deserialization of TimeType objects.

    This test function creates a time type object with the given parameters and
    then serializes it and deserializes it. Also prints it for visual inspection
    of the formatted output.

    Args:
        t_base (int): Time base for the new time type object
        t_context (int): Time context for the new time type object
        secs (int): Seconds value for the new time type object
        usecs (int): Seconds value for the new time type object
        should_err (int): True if error expected, else False

    Returns:
        True if test passed, False otherwise
    """
    val = TimeType(t_base, t_context, secs, usecs)

    buff = val.serialize()

    val2 = TimeType()
    val2.deserialize(buff, 0)

    assert val2.timeBase.value == t_base
    assert val2.timeContext == t_context
    assert val2.seconds == secs
    assert val2.useconds == usecs


def test_boolean_nominal():
    """ Tests the nominal cases of a BoolType """
    valid_values_test(BoolType, [True, False], 1)


def test_boolean_off_nominal():
    """ Tests the nominal cases of a BoolType """
    invalid_values_test(
        BoolType, filter(lambda item: not isinstance(item, bool), PYTHON_TESTABLE_TYPES)
    )


def test_int_types_nominal():
    """ Tests the integer types """
    for type_input, size in [(I8Type, 1), (I16Type, 2), (I32Type, 4), (I64Type, 8)]:
        total = pow(2, (size * 8) - 1)
        valid_values_test(type_input, [0, -1, 1, -total, total - 1], size)


def test_int_types_off_nominal():
    """ Tests the integer off nominal types """
    for type_input, size in [(I8Type, 1), (I16Type, 2), (I32Type, 4), (I64Type, 8)]:
        total = pow(2, (size * 8) - 1)
        invalid_values_test(
            type_input,
            filter(lambda item: not isinstance(item, int), PYTHON_TESTABLE_TYPES),
        )
        invalid_values_test(
            type_input, [-total - 1, total, -total * 35, total * 35], TypeRangeException
        )


def test_uint_types_nominal():
    """ Tests the integer types """
    for type_input, size in [(U8Type, 1), (U16Type, 2), (U32Type, 4), (U64Type, 8)]:
        max_int = pow(2, (size * 8)) - 1
        valid_values_test(type_input, [0, 1, max_int - 1, max_int], size)


def test_uint_types_off_nominal():
    """ Tests the integer off nominal types """
    for type_input, size in [(U8Type, 1), (U16Type, 2), (U32Type, 4), (U64Type, 8)]:
        max_int = pow(2, (size * 8)) - 1
        invalid_values_test(
            type_input,
            filter(lambda item: not isinstance(item, int), PYTHON_TESTABLE_TYPES),
        )
        invalid_values_test(
            type_input,
            [-1, -2, max_int + 1, max_int * 35, -max_int],
            TypeRangeException,
        )


def test_float_types_nominal():
    """ Tests the integer types """
    valid_values_test(F32Type, [0.31415000557899475, 0.0, -3.141590118408203], 4)
    valid_values_test(F64Type, [0.31415000557899475, 0.0, -3.141590118408203], 8)


def test_float_types_off_nominal():
    """ Tests the integer off nominal types """
    invalid_values_test(
        F32Type, filter(lambda item: not isinstance(item, float), PYTHON_TESTABLE_TYPES)
    )
    invalid_values_test(
        F64Type, filter(lambda item: not isinstance(item, float), PYTHON_TESTABLE_TYPES)
    )


def test_enum_type():
    """
    Tests the EnumType serialization and deserialization
    """
    members = {"MEMB1": 0, "MEMB2": 6, "MEMB3": 9}
    val1 = EnumType("SomeEnum", members, "MEMB3")
    buff = val1.serialize()
    val2 = EnumType("SomeEnum", members)
    val2.deserialize(buff, 0)
    assert val1.val == val2.val


def check_cloned_member_list(members1, members2):
    """ Check member list knowing direct compares don't work"""
    for tuple1, tuple2 in zip(members1, members2):
        assert tuple1[0] == tuple2[0], "Names do not match"
        assert tuple1[2] == tuple2[2], "Format strings do not match"
        assert tuple1[3] == tuple2[3], "Descriptions do not match"
        assert tuple1[1].val == tuple2[1].val, "Values don't match"


def test_serializable_type():
    """
    Tests the SerializableType serialization and deserialization
    """
    u32Mem = U32Type(1000000)
    stringMem = StringType("something to say")
    members = {"MEMB1": 0, "MEMB2": 6, "MEMB3": 9}
    enumMem = EnumType("SomeEnum", members, "MEMB3")
    memList = [
        ("mem1", u32Mem, ">i"),
        ("mem2", stringMem, ">H"),
        ("mem3", enumMem, ">i"),
    ]
    serType1 = SerializableType("ASerType", memList)
    buff = serType1.serialize()
    serType2 = SerializableType("ASerType", memList)
    serType2.deserialize(buff, 0)
    check_cloned_member_list(serType1.mem_list, serType2.mem_list)

    assert serType1.val == serType2.val

    i32Mem = I32Type(-1000000)
    stringMem = StringType("something else to say")
    members = {"MEMB1": 4, "MEMB2": 2, "MEMB3": 0}
    enumMem = EnumType("SomeEnum", members, "MEMB3")
    memList = [
        ("mem1", i32Mem, ">i"),
        ("mem2", stringMem, ">H"),
        ("mem3", enumMem, ">i"),
    ]
    serType1 = SerializableType("ASerType", memList)
    buff = serType1.serialize()
    serType2 = SerializableType("ASerType", memList)
    serType2.deserialize(buff, 0)
    check_cloned_member_list(serType1.mem_list, serType2.mem_list)

    value_dict = {"mem1": 3, "mem2": "abc 123", "mem3": "MEMB1"}
    serType1.val = value_dict
    assert serType1.val == value_dict
    mem_list = serType1.mem_list
    memList = [(a, b, c, None) for a, b, c in memList]
    check_cloned_member_list(mem_list, memList)

    serTypeEmpty = SerializableType("ASerType", [])
    assert serTypeEmpty.val == {}
    assert serTypeEmpty.mem_list == []


# def test_array_type():
#    """
#    Tests the ArrayType serialization and deserialization
#    """
#    extra_ctor_args = [("TestArray", (I32Type, 2, "I DON'T KNOW")), ("TestArray2", (U8Type, 4, "I DON'T KNOW")),
#               ("TestArray3", (StringType, 1, "I DON'T KNOW"))]
#    values = [[32, 1], [0, 1, 2, 3], ["one"]]
#    sizes = [8, 4, 3]
#
#    valid_values_test(ArrayType, values, sizes, extra_ctor_args)


def test_time_type():
    """
    Tests the TimeType serialization and deserialization
    """
    TIME_SIZE = 11

    in_no_err_list = [
        (TimeBase["TB_NONE"].value, 1, 100, 999999),
        (TimeBase["TB_PROC_TIME"].value, 0xFF, 1234567, 2952),
        (TimeBase["TB_WORKSTATION_TIME"].value, 8, 1529430215, 12),
        (TimeBase["TB_SC_TIME"].value, 231, 1344230277, 123456),
        (TimeBase["TB_FPGA_TIME"].value, 78, 10395, 24556),
        (TimeBase["TB_DONT_CARE"].value, 0xB3, 12390819, 12356),
    ]

    in_err_list = [
        (10, 58, 15345, 0),
        (TimeBase["TB_NONE"].value, 1, 3, -1),
        (TimeBase["TB_WORKSTATION_TIME"].value, 1, 700000, 1234567),
    ]

    val = TimeType()
    size = val.getSize()
    assert size == TIME_SIZE

    for (t_base, t_context, secs, usecs) in in_no_err_list:
        ser_deser_time_test(t_base, t_context, secs, usecs)

    for (t_base, t_context, secs, usecs) in in_err_list:
        with pytest.raises(TypeRangeException):
            ser_deser_time_test(t_base, t_context, secs, usecs)


class Dummy(BaseType):
    def serialize(self):
        return "serialized"

    def deserialize(self, data, offset):
        super(Dummy, self).deserialize(data, offset)
        return "deserialized"

    def getSize(self):
        return 0

    def to_jsonable(self):
        return {'name': 'dummy'}


def test_base_type():
    with pytest.raises(TypeError) as excinfo:
        BaseType()

    assert "Can't instantiate abstract class" in str(excinfo.value)

    with pytest.raises(TypeError) as excinfo2:
        ValueType()

    assert "Can't instantiate abstract class" in str(excinfo2.value)

    d = Dummy()
    assert d.serialize() == "serialized"
    assert d.getSize() == 0
    with pytest.raises(AbstractMethodException):
        # In the Dummy class above, the deserialize method
        # is set to call the super class, which is just the
        # raw abstract method, which is the only way to
        # raise an `AbstractMethodException`.
        d.deserialize("a", 0)



"""
fprime.util.__main__:

This acts as the main entry point for the fprime.util module. This allows users to run fprime.util as an entry point.
This will include the build_helper scripts and run them.
"""
import sys

import fprime.util.build_helper


def main():
    """ Run wrapper, to point a console_script at """
    return fprime.util.build_helper.utility_entry(args=sys.argv[1:])


if __name__ == "__main__":
    sys.exit(main())



"""
(test) fprime.build:

Tests the current F prime build module. Ensures that the selected singleton defines the minimum set of functions and
that they function as expected.

@author mstarch
"""
import os
import shutil
import tempfile
import pathlib

import pytest

import fprime.fbuild.cmake
import fprime.fbuild.builder


def get_cmake_builder():
    """Gets a CMake builder for these tests

    Returns:
        CMakeBuilder for testing
    """
    return fprime.fbuild.cmake.CMakeHandler()


def get_data_dir():
    """
    Gets directory containing test-data specific to the builder being tested. This will enable new implementers, should
    there be any, to implement their own build-directory structure.

    :return:
    """
    if type(get_cmake_builder()) == fprime.fbuild.cmake.CMakeHandler:
        return os.path.join(os.path.dirname(__file__), "cmake-data")
    raise Exception(
        "Test data directory not setup for {} builder class".format(
            type(get_cmake_builder())
        )
    )


def test_hash_finder():
    """
    Tests that the hash finder works given a known builds.
    """
    local = pathlib.Path(os.path.dirname(__file__))
    dep_dir = local / "cmake-data" / "testbuild"
    builder = fprime.fbuild.builder.Build(
        fprime.fbuild.builder.BuildType.BUILD_NORMAL, dep_dir
    )
    builder.load(local, build_dir=dep_dir / "build-fprime-automatic-native")

    assert builder.find_hashed_file(0xDEADBEEF) == ["Abc: 0xdeadbeef\n"]
    assert builder.find_hashed_file(0xC0DEC0DE) == ["HJK: 0xc0dec0de\n"]


def test_needed_functions():
    """
    Test the needed functions for the given builder. This will ensure that the public interface to the builder is
    implemented as expected.
    """
    needed_funcs = [
        "get_include_info",
        "execute_known_target",
        "get_include_locations",
        "get_fprime_configuration",
    ]
    for func in needed_funcs:
        assert hasattr(get_cmake_builder(), func)


def test_get_fprime_configuration():
    """
    Tests the given fprime configuration fetcher. Required for other portion of the system.
    """
    configs = fprime.fbuild.cmake.CMakeHandler.CMAKE_LOCATION_FIELDS
    test_data = {
        "grand-unified": (
            "/home/user11/fprime/Ref/..",
            None,
            "/home/user11/fprime/Ref/..",
        ),
        "subdir": (
            "/home/user11/Proj",
            "/home/user11/Proj/lib1;/home/user11/Proj/lib2",
            "/home/user11/Proj/fprime",
        ),
        "external": ("/home/user11/Proj", "/opt/lib1;/opt/lib2", "/opt/fprime"),
    }
    for key in test_data.keys():
        build_dir = os.path.join(get_data_dir(), key)
        # Test all path, truth pairs
        values = get_cmake_builder().get_fprime_configuration(configs, build_dir)
        assert values == test_data[key]


def test_get_include_locations():
    """
    Test all the include locations. This will ensure that values are properly read from a cache listing. This will
    support various other portions of the system, so debug here firest.
    """
    test_data = {
        "grand-unified": ["/home/user11/fprime"],
        "subdir": [
            "/home/user11/Proj",
            "/home/user11/Proj/lib1",
            "/home/user11/Proj/lib2",
            "/home/user11/Proj/fprime",
        ],
        "external": ["/home/user11/Proj", "/opt/lib1", "/opt/lib2", "/opt/fprime"],
    }
    for key in test_data.keys():
        build_dir = os.path.join(get_data_dir(), key)
        paths = list(get_cmake_builder().get_include_locations(build_dir))
        assert paths == test_data[key]


def test_get_include_info():
    """
    Tests that the include root function gets the expected value based on the path and build-directory setups.
    """
    # Test data setup, format build_dir to tuples of path, expected result. Note: None means expect Orphan exception
    test_data = {
        "grand-unified": [
            (
                "/home/user11/fprime/Svc/SomeComp1",
                ("Svc/SomeComp1", "/home/user11/fprime"),
            ),
            (
                "/home/user11/fprime/Ref/SomeComp2",
                ("Ref/SomeComp2", "/home/user11/fprime"),
            ),
            (
                "/home/user11/fprime/Ref/SomeComp2/../SomeComp1",
                ("Ref/SomeComp1", "/home/user11/fprime"),
            ),
            ("/home/user11/external-sw/NachoDeploy/SomeComp3", None),
        ],
        "subdir": [
            (
                "/home/user11/Proj/fprime/Svc/SomeComp1",
                ("Svc/SomeComp1", "/home/user11/Proj/fprime"),
            ),
            # An insidious case where the path contains a common prefix as another possible location without being an
            # exact directory.  Notice /home/user11/Proj/fprime-something/Comps/SomeComp1 contains a non-exact common
            # prefix with the directory /home/user11/Proj/fprime. i.e. /home/user11/Proj/fprime is a prefix but
            # /home/user11/Proj/fprime/ is not.
            (
                "/home/user11/Proj/fprime-something/Comps/SomeComp1",
                ("fprime-something/Comps/SomeComp1", "/home/user11/Proj"),
            ),
            ("/home/user11/Proj/Ref/SomeComp2", ("Ref/SomeComp2", "/home/user11/Proj")),
            (
                "/home/user11/Proj/Ref/SomeComp2/../../fprime/Svc/SomeComp1",
                ("Svc/SomeComp1", "/home/user11/Proj/fprime"),
            ),
            ("/home/user11/external-sw/NachoDeploy/SomeComp3", None),
        ],
        "external": [
            ("/opt/fprime/Svc/SomeComp1", ("Svc/SomeComp1", "/opt/fprime")),
            ("/home/user11/Proj/Ref/SomeComp2", ("Ref/SomeComp2", "/home/user11/Proj")),
            ("/opt/something/else/external-sw/NachoDeploy/SomeComp3", None),
        ],
    }
    # Run through all the above data look for matching answers
    for key in test_data.keys():
        build_dir = os.path.join(get_data_dir(), key)
        # Test all path, truth pairs
        for path, truth in test_data.get(key):
            if truth is None:
                with pytest.raises(fprime.fbuild.cmake.CMakeOrphanException):
                    value = get_cmake_builder().get_include_info(path, build_dir)
            else:
                value = get_cmake_builder().get_include_info(path, build_dir)
                assert value == truth


def test_find_nearest_deployment():
    """
    This will test the ability for the system to detect valid deployment directories
    """

    test_dir = pathlib.Path(get_data_dir())
    test_data = [
        ("testbuild/subdir1/subdir2/subdir3", "testbuild/subdir1/"),
        ("testbuild/subdir1/subdir2", "testbuild/subdir1/"),
        ("testbuild/", "testbuild/"),
        ("/nonexistent/dirone/someotherpath", None),
    ]
    for path, truth in test_data:
        path = test_dir / path
        if truth is not None:
            truth = test_dir / truth
            value = fprime.fbuild.builder.Build.find_nearest_deployment(path)
            assert value == truth
        else:
            with pytest.raises(fprime.fbuild.builder.UnableToDetectDeploymentException):
                fprime.fbuild.builder.Build.find_nearest_deployment(path)


def test_generate():
    """
    Generate a directory and ensure that it works via CMake. This tests the common setup flags for normal and testing
    builds. Will also test the right errors are raised.
    """
    rms = []
    try:
        test_flags = [{}, {"CMAKE_BUILD_TYPE": "Testing"}]
        # Build Ref with flags
        path = os.path.join(
            os.path.dirname(__file__), "..", "..", "..", "..", "..", "Ref"
        )
        for flags in test_flags:
            # Create a temp directory and register its deletion at the end of the program run
            tempdir = tempfile.mkdtemp()
            rms.append(tempdir)
            get_cmake_builder().generate_build(path, tempdir, flags, ignore_output=True)
        # Expect errors for this step
        path = "/nopath/somesuch/nothing"
        with pytest.raises(fprime.fbuild.cmake.CMakeProjectException):
            tempdir = tempfile.mkdtemp()
            rms.append(tempdir)
            get_cmake_builder().generate_build(path, tempdir, ignore_output=True)
    # Clean-Up all the directories made
    finally:
        for rmd in rms:
            shutil.rmtree(rmd, ignore_errors=True)


def test_targets():
    """
    Test standard targets for the build.
    """
    # Build Ref with flags
    tempdir = None
    try:
        fprime_root = os.path.join(
            os.path.dirname(__file__), "..", "..", "..", "..", ".."
        )
        # Create a temp directory and register its deletion at the end of the program run
        tempdir = tempfile.mkdtemp()
        get_cmake_builder().generate_build(
            os.path.join(fprime_root, "Ref"), tempdir, {"CMAKE_BUILD_TYPE": "Testing"}
        )
        test_data = [
            (os.path.join(fprime_root, "Ref"), ""),
            (os.path.join(fprime_root, "Svc", "CmdDispatcher"), ""),
            (os.path.join(fprime_root, "Svc", "CmdDispatcher"), "ut_exe"),
            (os.path.join(fprime_root, "Svc", "CmdDispatcher"), "check"),
        ]
        # Loop over all directories and target pairs ensuing things work
        for path, target in test_data:
            get_cmake_builder().execute_known_target(target, tempdir, path)
        test_data = [
            (os.path.join(fprime_root, "Svc", "CmdDispatcher"), "nontarget1"),
            (os.path.join(fprime_root, "Svc", "CmdDispatcher3Not"), ""),
        ]
        # Loop over all paths and target pairs looking for expected Exceptions
        for path, target in test_data:
            with pytest.raises(fprime.fbuild.cmake.CMakeException):
                get_cmake_builder().execute_known_target(target, tempdir, path)
    # Clean up when all done
    finally:
        if tempdir is not None:
            shutil.rmtree(tempdir, ignore_errors=True)



"""
Supplies high-level build functions to the greater fprime helper CLI. This maps from user command space to the specific
build system handler underneath.
"""
import os
import re
import functools
from abc import ABC
from enum import Enum
from pathlib import Path
from typing import Iterable, List, Set, Union

from fprime.common.error import FprimeException
from fprime.fbuild.settings import IniSettings
from fprime.fbuild.cmake import CMakeHandler, CMakeException


class BuildType(Enum):
    """
    An enumeration used to represent the various build types used to build fprime. These types can support different
    types of targets underneath. i.e. the unit-test build may build unit test executables.
    """

    """ Normal build normal binaries for a deployment mapping to CMake 'Release'"""  # pylint: disable=W0105
    BUILD_NORMAL = (0,)
    """ Testing build allowing unit testing mapping to CMake 'Testing'"""  # pylint: disable=W0105
    BUILD_TESTING = 1

    def get_suffix(self):
        """ Get the suffix of a directory supporting this build """
        if self == BuildType.BUILD_NORMAL:
            return ""
        if self == BuildType.BUILD_TESTING:
            return "-ut"
        raise InvalidBuildTypeException(
            "{} is not a supported build type".format(self.name)
        )

    def get_cmake_build_type(self):
        """ Get the suffix of a directory supporting this build """
        if self == BuildType.BUILD_NORMAL:
            return "Release"
        if self == BuildType.BUILD_TESTING:
            return "Testing"
        raise InvalidBuildTypeException(
            "{} is not a supported build type".format(self.name)
        )


class Target(ABC):
    """Generic build target base class

    A target can be specified by the user using a mnemonic and flags. The mnemonic is the command typed in by the user,
    and the flags allow the user to remember fewer mnemonics by changing the build target using a modifier. Each build
    target is available in certain build types.

    Targets can be global, using the GlobalTarget base class. Global targets don't use contextual information to modify
    the target, but apply to the whole deployment. Note: global targets are also engaged at the deployment level should
    that be the context.

    Targets may also be local. These targets use context information to figure out what to build. This allows for one
    target to represent a class of targets. i.e. build can be used as a local target to build any given sub directory.
    """

    def __init__(
        self,
        mnemonic: str,
        desc: str,
        build_type: BuildType = None,
        flags: set = None,
        cmake: str = None,
    ):
        """Constructs a build target

        Args:
            mnemonic:    mnemonic used to engage build targets. Is not unique, but mnemonic + flags must be.
            desc:        help description of this build target
            build_types: supported build types for target. Defaults to [BuildType.BUILD_NORMAL, BuildType.BUILD_TESTING]
            flags:       flags used to uniquely identify build targets who share logical mnemonics. Defaults to None.
            cmake:       cmake target override to handle oddly named cmake targets
        """
        self.mnemonic = mnemonic
        self.desc = desc
        self.build_type = (
            build_type if build_type is not None else BuildType.BUILD_NORMAL
        )
        self.flags = flags if flags is not None else set()
        self.cmake_target = cmake if cmake is not None else mnemonic

    def __str__(self):
        """ Makes this target into a string """
        return self.config_string(self.mnemonic, self.flags)

    @staticmethod
    def config_string(mnemonic, flags):
        """Converts a mnemonic and set of flags to string

        Args:
            mnemonic: mnemonic of the target
            flags: sset of flags to pair with mnemonic
        Returns:
            string of format "mnemonic --flag1 --flag2 ..."
        """
        flag_string = " ".join(["--{}".format(flag) for flag in flags])
        flag_string = "" if flag_string == "" else " " + flag_string
        return "{}{}".format(mnemonic, flag_string)

    @classmethod
    def get_all_possible_flags(cls) -> Set[str]:
        """Gets list of all targets' flags used

        Returns:
            List of targets's supported by the system
        """
        return functools.reduce(
            lambda agg, item: agg.union(item.flags), cls.get_all_targets(), set()
        )

    @classmethod
    def get_all_targets(cls) -> List["Target"]:
        """Gets list of all targets registered

        Returns:
            List of targets supported by the system
        """
        return BUILD_TARGETS

    @classmethod
    def get_target(cls, mnemonic: str, flags: Set[str]) -> "Target":
        """Gets the actual build target given the parsed namespace

        Using the global list of build targets and the flags supplied to the namespace, attempt to determine which build
        targets can be used. If more than one are found, then generate exception.

        Args:
            mnemonic: mnemonic of command to look for
            flags:    flags to narrow down target

        Returns:
            single matching target
        """
        matching = []
        for target in cls.get_all_targets():
            if target.mnemonic == mnemonic and flags == target.flags:
                matching.append(target)
        if not matching:
            raise NoSuchTargetException(
                "Could not find target '{}'".format(cls.config_string(mnemonic, flags))
            )
        assert len(matching) == 1, "Conflicting targets specified in code"
        return matching[0]


class GlobalTarget(Target):
    """Represents a global build target

    Build targets are global if they do not apply to a specific directory, but rather to a full deployment.
    """


class LocalTarget(Target):
    """Represents a local build target

    Build targets are local if they do  apply to a specific directory whose context drives the build setup.
    """


class Build:
    """Represents a build configuration

    Builds in F´ consist of a build type (normal, testing), a deployment directory, a set of settings, and a target
    platform. These are tracked as part of this Build class. This helps setup a build cache directory, load default
    settings, and track what type of build is being run.

    BuildType represents the type of build as explained in that enum type.
    Deployments are an individual build of fprime, and should define the CMakeLists.txt file as a child of this
    directory. A default settings.ini file may be found here.
    Platforms represent the target hardware to build from. This is translated to the CMake toolchain file.

    After creation, a user must use invent to handle new builds (e.g. during the generation step), or load to load a
    previously generated build.

    Examples:
        To use in generation run the following code.

        build = Build(BuildType.BUILD_NORMAL, path/to/deployment)
        build.invent("raspberrypi")

        To use at any step after generation:

        build = Build(BuildType.BUILD_NORMAL, path/to/deployment)
        build.load()
    """

    VALID_CMAKE_LIST = re.compile(r"^\s*project\(.*\)", re.MULTILINE)
    CMAKE_DEFAULT_BUILD_NAME = "build-fprime-automatic-{platform}{suffix}"

    def __init__(self, build_type: BuildType, deployment: Path, verbose: bool = False):
        """Constructs a build object from its constituent parts

        Args:
            build_type: member of the enum BuildType specifying fprime build type
            deployment: path to deployment that this build represents
        """
        self.build_type = build_type
        self.deployment = deployment
        self.settings = None
        self.platform = None
        self.build_dir = None
        self.cmake = CMakeHandler()
        self.cmake.set_verbose(verbose)

    def invent(self, cwd: Path, platform: str = None, build_dir: Path = None):
        """Invents a build path from a given platform

        Sets this build up as a new build that would be used as as part of a generate step. This directory must not
        already exist. If platform is None, a default will be chosen from the settings.ini file. If the settings.ini
        file does not exist, or does not specify a default_toolchain, then "native" will be used. Settings are loaded in
        this step for further uses of this build.

        build_dir is used to specify an exact build directory to use as part of this step. This allows directories to be
        specified by the caller, but is typically not used.

        Args:
            platform:   name of platform to build against. None will use default from settings.ini or without this
                        setting, "native". Defaults to None.
            build_dir:  explicitly sets the build path to allow for user override of default

        Raises:
            InvalidBuildCacheException: a build cache already exists as it should not
        """
        self.__setup_default(cwd, platform, build_dir)
        if self.build_dir.exists():
            raise InvalidBuildCacheException(
                "{} already exists.".format(self.build_dir)
            )

    def load(self, cwd: Path, platform: str = None, build_dir: Path = None):
        """Load an existing build cache

        Sets this build up from an existing build cache. This can be used after a previous run that has generated a
        build cache in order to prepare for other build steps.

        Args:
            platform:   name of platform to build against. None will use default from settings.ini or without this
                        setting, "native". Defaults to None.
            build_dir:  explicitly sets the build path to allow for user override of default

        Raises:
            InvalidBuildCacheException: the build cache does not exist as it must
        """
        self.__setup_default(cwd, platform, build_dir)
        if (
            not self.build_dir.exists()
            or not (self.build_dir / "CMakeCache.txt").exists()
        ):
            gen_args = " --ut" if self.build_type == BuildType.BUILD_TESTING else ""
            raise InvalidBuildCacheException(
                "'{}' is not a valid build cache. Generate this build cache with 'fprime-util generate{}'".format(
                    self.build_dir, gen_args
                )
            )

    def get_settings(
        self, setting: Union[str, Iterable[str]], default: Union[str, Iterable[str]]
    ) -> Union[str, Iterable[str]]:
        """Fetches settings in the settings file

        Reads settings loaded from the settings file and returns them to the caller. If a single string is submitted,
        then a single string is returned. If a list of strings is submitted a list is returned. default provides default
        values to supply in the case that a setting is unavailable.

        Args:
            setting: a string or set of string settings to return
            default: a string or set of string settings to return if no setting is found

        Returns:
            a single string setting or a list of string settings to match request with defaults subbed ins
        """
        if isinstance(setting, str):
            return self.settings.get(setting, default)
        return [self.get_settings(req, back) for req, back in zip(setting, default)]

    def find_hashed_file(self, hash_value: int) -> List[str]:
        """Retrieves the file associated with a hash

        In order to reduce space and memory footprint, filenames are associated with hashes automatically as part of the
        build. This function will retrieve the file name given a has integer.

        Args:
            hash_value: hash number to lookup

        Returns:
            stored file path(s) associated with hash
        """
        hashes_file = self.build_dir / "hashes.txt"
        if not hashes_file.exists():
            raise InvalidBuildCacheException(
                "Failed to find {}, was the build generated.".format(hashes_file)
            )
        with open(hashes_file) as file_handle:
            lines = filter(
                lambda line: "{:x}".format(hash_value) in line, file_handle.readlines()
            )
        return list(lines)

    def get_build_cache(self) -> Path:
        """Generates build cache path for this build

        Generates the build path for this build. This will expect a valid build path to exist unless validate is
        specified as false. A valid build cache has been created from the generate step, and thus when using this call
        as part of the generate step, validate should be set to false.

        Returns:
            Path to a build cache directory

        """
        return self.deployment / Build.CMAKE_DEFAULT_BUILD_NAME.format(
            platform=self.platform, suffix=self.build_type.get_suffix()
        )

    def get_build_info(self, context: Path) -> dict:
        """Constructs an informational packet about this build

        Constructs a packet that allows for users to get meta-build information. This includes: location of build, file
        and other constructs, available make targets, and other items.

        Args:
            context: contextual path to list various information about the build

        Returns:
            Build information dictionary
        """
        valid_cmake_targets = self.cmake.get_available_targets(
            str(self.build_dir), context
        )
        local_targets = [
            target
            for target in BUILD_TARGETS
            if target.cmake_target in valid_cmake_targets
            and isinstance(target, LocalTarget)
        ]
        global_targets = [
            target for target in BUILD_TARGETS if isinstance(target, GlobalTarget)
        ]

        relative_path = self.cmake.get_project_relative_path(
            str(context), self.build_dir
        )

        for possible in ["F-Prime", "."]:
            auto_location = self.build_dir / possible / relative_path
            if auto_location.exists():
                break
        else:
            auto_location = None
        return {
            "local_targets": local_targets,
            "global_targets": global_targets,
            "auto_location": auto_location,
        }

    def find_toolchain(self):
        """Locates a toolchain file in know locations

        Finds a toolchain for the given platform.  Searches in known locations for the toolchain, and compares against F
        prime provided toolchains, toolchains in libraries, and toolchains provided by project.

        Returns:
            path to CMake toolchain file or None to use builtin
        """
        assert (
            self.platform != "default"
        ), "Default toolchain should have been decided already"
        toolchain_locations = self.get_settings(
            ["framework_path", "project_root"], [None, self.deployment]
        )
        toolchain_locations += self.get_settings("library_locations", [])

        # If toolchain is the native target, this is supplied by CMake and we exit here.
        if self.platform == "native":
            return None
        # Otherwise, find locations of toolchain files using the specified locations from settings.
        toolchains_paths = [
            os.path.join(loc, "cmake", "toolchain", self.platform + ".cmake")
            for loc in toolchain_locations
            if loc is not None
        ]
        # Create a deduplicated set of toolchains
        toolchains = list(
            {
                toolchain_path
                for toolchain_path in toolchains_paths
                if os.path.exists(toolchain_path)
            }
        )
        if not toolchains:
            raise NoSuchToolchainException(
                "Could not find toolchain file for {} at any of: {}".format(
                    self.platform, " ".join(toolchains_paths)
                )
            )
        if len(toolchains) > 1:
            raise AmbiguousToolchainException(
                "Found conflicting toolchain files for {} at: {}".format(
                    self.platform, " ".join(toolchains)
                )
            )
        return toolchains[0]

    def get_cmake_args(self) -> dict:
        """Generates CMake arguments from deployment settings (settings.ini file)

        Returns:
            A dictionary of cmake settings
        """
        needed = [
            ("FPRIME_FRAMEWORK_PATH", "framework_path"),
            ("FPRIME_LIBRARY_LOCATIONS", "library_locations"),
            ("FPRIME_PROJECT_ROOT", "project_root"),
            ("FPRIME_SETTINGS_FILE", "settings_file"),
            ("FPRIME_ENVIRONMENT_FILE", "environment_file"),
            ("FPRIME_AC_CONSTANTS_FILE", "ac_constants"),
            ("FPRIME_CONFIG_DIR", "config_dir"),
            ("FPRIME_INSTALL_DEST", "install_dest"),
        ]
        cmake_args = {
            cache: self.get_settings(setting, None)
            for cache, setting in needed
            if self.get_settings(setting, None) is not None
        }

        if "FPRIME_LIBRARY_LOCATIONS" in cmake_args:
            cmake_args["FPRIME_LIBRARY_LOCATIONS"] = ";".join(
                cmake_args["FPRIME_LIBRARY_LOCATIONS"]
            )

        cmake_args.update({"CMAKE_BUILD_TYPE": self.build_type.get_cmake_build_type()})
        return cmake_args

    def execute(self, target: Target, context: Path, make_args: dict):
        """Execute a build target

        Executes a target within the build system. This will execute the target by calling into the make system. Context
        is supplied such that the system can match local targets to the global target list.

        Args:
            target: target to run
            context: context path for local targets
            make_args: make system arguments directly supplied
        """
        self.cmake.execute_known_target(
            target.cmake_target,
            self.build_dir,
            context.absolute(),
            cmake_args=self.get_cmake_args(),
            make_args=make_args,
            top_target=isinstance(target, GlobalTarget),
            environment=self.settings.get("environment", None),
        )

    def generate(self, cmake_args):
        """Generates a build given CMake arguments

        This will run a generate step of the cmake build process. This will take in any argument used/passed to CMake.

        Args:
            cmake_args: cmake arguments to pass into the generate step
        """
        try:
            self.cmake.generate_build(
                self.deployment,
                self.build_dir,
                {**cmake_args, **self.get_cmake_args()},
                environment=self.settings.get("environment", None),
            )
        except CMakeException as cexc:
            raise GenerateException(str(cexc)) from cexc

    def purge(self):
        """ Purge a build cache directory """
        self.cmake.purge(self.build_dir)

    def purge_install(self):
        """ Purge the install directory """
        assert "install_dest" in self.settings, "install_dest not present in settings"
        self.cmake.purge(self.settings["install_dest"])

    def install_dest_exists(self) -> Path:
        """ Check if the install destination exists and returns the path if it does """
        assert "install_dest" in self.settings, "install_dest not present in settings"
        path = Path(self.settings["install_dest"])
        return path if path.exists() else None

    @staticmethod
    def find_nearest_deployment(path: Path) -> Path:
        """Recurse up the directory stack looking for a valid deployment

        Recurse up the directory tree from the given path, looking for a deployment definition directory. This means it
        defines a CMakeLists.txt with a project call. This finds where the automatic build directories are allowed to
        exist.

        Notes:
            This replaced the former build directory recursive detector as that can "slip" past a deployment should the
            build directory not be generated yet.

        Returns;
            Path to the nearest deployment directory searching up the directory tree

        Raises;
            UnableToDetectDeploymentException: was unable to detect a deployment directory
        """
        full_path = path.resolve()
        list_file = full_path / "CMakeLists.txt"
        if not full_path.parents:
            raise UnableToDetectDeploymentException()
        if list_file.exists():
            with open(list_file) as file_handle:
                text = file_handle.read()
            if Build.VALID_CMAKE_LIST.search(text):
                return full_path
        return Build.find_nearest_deployment(full_path.parent)

    def __setup_default(self, cwd: Path, platform: str = None, build_dir: Path = None):
        """Sets up default build

        Sets this build up before determining if it is a pre-generated, or post-generated build.

        build_dir is used to specify an exact build directory to use as part of this step. This allows directories to be
        specified by the caller, but is typically not used.

        Args:
            platform:   name of platform to build against. None will use default from settings.ini or without this
                        setting, "native". Defaults to None.
            build_dir:  explicitly sets the build path to allow for user override of default
        """
        assert self.settings is None, "Already setup it is invalid to re-setup"
        assert self.platform is None, "Already setup it is invalid to re-setup"
        assert self.build_dir is None, "Already setup it is invalid to re-setup"

        self.settings = IniSettings.load(self.deployment / "settings.ini", cwd)

        if platform is not None and platform != "default":
            self.platform = platform
        elif self.build_type == BuildType.BUILD_TESTING:
            self.platform = self.settings.get("default_ut_toolchain", "native")
        else:
            self.platform = self.settings.get("default_toolchain", "native")

        self.build_dir = build_dir if build_dir is not None else self.get_build_cache()


class GenerateException(FprimeException):
    """ An exception indicating generate has failed and the user may need to respond """


class InvalidBuildTypeException(FprimeException):
    """ An exception indicating a build type do not exit """


class InvalidBuildCacheException(FprimeException):
    """ An exception indicating a build cache """


class UnableToDetectDeploymentException(FprimeException):
    """ An exception indicating a build cache """


class NoSuchTargetException(FprimeException):
    """ Could not find a matching build target """


class NoSuchToolchainException(FprimeException):
    """ Could not find a matching build target """


class AmbiguousToolchainException(FprimeException):
    """ Could not find a matching build target """


""" Defined set of build targets available to the system"""  # pylint: disable=W0105
BUILD_TARGETS = [
    # Various "build" target
    LocalTarget("build", "Build components, ports, and deployments", cmake=""),
    GlobalTarget("build", "Build all deployment targets", flags={"all"}, cmake="all"),
    LocalTarget(
        "build",
        "Build unit tests",
        build_type=BuildType.BUILD_TESTING,
        flags={"ut"},
        cmake="ut_exe",
    ),
    # Implementation targets
    LocalTarget("impl", "Generate implementation template files"),
    LocalTarget("impl", "Generate unit test files", flags={"ut"}, cmake="testimpl"),
    # Check targets and unittest targets
    LocalTarget("check", "Run unit tests", build_type=BuildType.BUILD_TESTING),
    LocalTarget(
        "check",
        "Run unit tests with memory checking",
        build_type=BuildType.BUILD_TESTING,
        flags={"leak"},
        cmake="check_leak",
    ),
    LocalTarget(
        "check",
        "Run unit tests with code coverage",
        build_type=BuildType.BUILD_TESTING,
        flags={"coverage"},
        cmake="coverage",
    ),
    GlobalTarget(
        "check",
        "Run all deployment unit tests",
        build_type=BuildType.BUILD_TESTING,
        flags={"all"},
    ),
    GlobalTarget(
        "check",
        "Run all deployment unit tests with memory checking",
        build_type=BuildType.BUILD_TESTING,
        flags={"all", "leak"},
        cmake="check_leak",
    ),
    GlobalTarget(
        "check",
        "Run all deployment unit tests with code coverage",
        build_type=BuildType.BUILD_TESTING,
        flags={"all", "coverage"},
        cmake="coverage",
    ),
]



""" Generic representation of autocoded array types

Created on May 29, 2020
@author: jishii
"""
import copy

from .type_base import ValueType
from .type_exceptions import (
    ArrayLengthException,
    NotInitializedException,
    TypeMismatchException,
)


class ArrayType(ValueType):
    """Generic fixed-size array type representation.

    Represents a custom named type of a fixed number of like members, each of which are other types in the system.
    """

    def __init__(self, typename, config_info, val=None):
        """Constructs a new array type.

        Args:
            typename: name of this array type
            config_info: (type, size, format string) information for array
            val: (optional) list of values to assign to array
        """
        super().__init__()
        if not isinstance(typename, str):
            raise TypeMismatchException(str, type(typename))
        self.__val = None
        self.__typename = typename
        self.__arr_type, self.__arr_size, self.__arr_format = config_info
        # Set value only if it is a valid, non-empty list
        if not val:
            return
        self.val = val

    def validate(self, val):
        """ Validates the values of the array """
        size = self.__arr_size
        if len(val) != size:
            raise ArrayLengthException(self.__arr_type, size, len(val))
        for i in range(self.__arr_size):
            if not isinstance(val[i], type(self.__arr_type)):
                raise TypeMismatchException(type(self.__arr_type), type(val[i]))

    @property
    def val(self) -> list:
        """
        The .val property typically returns the python-native type. This the python native type closes to a serializable
        without generating full classes would be a dictionary (anonymous object). This returns such an object.

        :return dictionary of member names to python values of member keys
        """
        return [item.val for item in self.__val]

    @val.setter
    def val(self, val: list):
        """
        The .val property typically returns the python-native type. This the python native type closes to a serializable
        without generating full classes would be a dictionary (anonymous object). This takes such an object and sets the
        member val list from it.

        :param val: dictionary containing python types to key names. This
        """
        items = []
        for item in val:
            cloned = copy.deepcopy(self.arr_type)
            cloned.val = item
            items.append(cloned)
        self.__val = items

    def to_jsonable(self):
        """
        JSONable type
        """
        members = {
            "name": self.__typename,
            "type": self.__typename,
            "size": self.__arr_size,
            "format": self.__arr_format,
            "values": None
            if self.__val is None
            else [member.to_jsonable() for member in self.__val],
        }
        return members

    def serialize(self):
        """ Serialize the array by serializing the elements one by one """
        if self.val is None:
            raise NotInitializedException(type(self))
        return b"".join([item.serialize() for item in self.val])

    def deserialize(self, data, offset):
        """ Deserialize the members of the array """
        values = []
        for i in range(self.__arr_size):
            item = copy.deepcopy(self.arr_type)
            item.deserialize(data, offset + i * item.getSize())
            values.append(item.val)
        self.val = values

    @property
    def arr_type(self):
        """ Property representing the size of the array """
        return self.__arr_type

    @property
    def arr_size(self):
        """ Property representing the number of elements of the array """
        return self.__arr_size

    @property
    def arr_format(self):
        """ Property representing the format string of an item in the array """
        return self.__arr_format

    def getSize(self):
        """ Return the size of the array """
        return self.arr_type.getSize() * self.arr_size



"""
Created on Dec 18, 2014

@author: reder
Replaced type base class with decorators
"""
import abc
import struct

from .type_exceptions import AbstractMethodException


class BaseType(abc.ABC):
    """
    An abstract base defining the methods supported by all base classes.
    """

    @abc.abstractmethod
    def serialize(self):
        """
        Serializes the current object type.
        """
        raise AbstractMethodException("serialize")

    @abc.abstractmethod
    def deserialize(self, data, offset):
        """
        AbstractDeserialize interface
        """
        raise AbstractMethodException("deserialize")

    @abc.abstractmethod
    def getSize(self):
        """
        Abstract getSize interface
        """
        raise AbstractMethodException("getSize")

    def __repr__(self):
        """ Produces a string representation of a given type """
        return self.__class__.__name__.replace("Type", "")

    @abc.abstractmethod
    def to_jsonable(self):
        """
        Converts this type to a JSON serializable object
        """
        raise AbstractMethodException("to_jsonable")


class ValueType(BaseType):
    """
    An abstract base type used to represent a single value. This defines the value property, allowing for setting and
    reading from the .val member.
    """

    def __init__(self, val=None):
        """ Defines the single value """
        self.__val = None
        # Run full setter
        if val is not None:
            self.val = val

    @abc.abstractmethod
    def validate(self, val):
        """
        Checks the val for validity with respect to the current type. This will raise TypeMismatchException when the
        validation fails of the val's type fails. It will raise TypeRangeException when val is out of range.

        :param val: value to validate
        :raises TypeMismatchException: value has incorrect type, TypeRangeException: val is out of range
        """

    @property
    def val(self):
        """ Getter for .val """
        return self.__val

    @val.setter
    def val(self, val):
        """ Setter for .val calls validate internally """
        self.validate(val)
        self.__val = val

    def to_jsonable(self):
        """
        Converts this type to a JSON serializable object
        """
        return {"value": self.val, "type": str(self)}


#
#
def showBytes(byteBuffer):
    """
    Routine to show bytes in buffer for testing.
    """
    print("Byte buffer size: %d" % len(byteBuffer))
    for entry in range(0, len(byteBuffer)):
        print(
            "Byte %d: 0x%02X (%c)"
            % (
                entry,
                struct.unpack("B", bytes([byteBuffer[entry]]))[0],
                struct.unpack("B", bytes([byteBuffer[entry]]))[0],
            )
        )



"""
Created on Dec 18, 2014
@author: tcanham, reder
"""
import struct

from .type_base import ValueType
from .type_exceptions import (
    DeserializeException,
    EnumMismatchException,
    NotInitializedException,
    TypeMismatchException,
    TypeRangeException,
)


class EnumType(ValueType):
    """
    Representation of the ENUM type.

    The enumeration takes a dictionary that stores the enumeration members
    and current value as a string. The member values will have to be computed
    containing code based on C enum rules
    """

    def __init__(self, typename="", enum_dict=None, val=None):
        """
        Constructor

        :param typename: name of the enumeration type
        :param enum_dict: dictionary of value to integer representation
        :param val: value of the enumeration
        """
        super().__init__()
        if not isinstance(typename, str):
            raise TypeMismatchException(str, type(val))
        self.__typename = typename
        # Setup the enum dictionary
        if enum_dict is None:
            enum_dict = {"UNDEFINED": 0}
        # Check if the enum dict is an instance of dictionary
        self.__enum_dict = enum_dict
        # Set val to undefined if not set
        if val is None:
            val = "UNDEFINED"
        self.val = val

    def validate(self, val):
        """ Validate the value passed into the enumeration """
        if not isinstance(self.enum_dict(), dict):
            raise TypeMismatchException(dict, type(self.enum_dict()))
        for member in self.keys():
            if not isinstance(member, str):
                raise TypeMismatchException(str, type(member))
            elif not isinstance(self.enum_dict()[member], int):
                raise TypeMismatchException(int, self.enum_dict()[member])
        if val != "UNDEFINED" and val not in self.keys():
            raise EnumMismatchException(self.__typename, val)

    def keys(self):
        """
        Return all the enum key values.
        """
        return list(self.enum_dict().keys())

    def typename(self):
        return self.__typename

    def enum_dict(self):
        return self.__enum_dict

    def serialize(self):
        """
        Serialize the enumeration type using an int type
        """
        # for enums, take the string value and convert it to
        # the numeric equivalent
        if self.val is None:
            raise NotInitializedException(type(self))
        return struct.pack(">i", self.enum_dict()[self.val])

    def deserialize(self, data, offset):
        """
        Deserialize the enumeration using an int type
        """
        try:
            int_val = struct.unpack_from(">i", data, offset)[0]
        except:
            raise DeserializeException(
                "Could not deserialize enum value. Needed: {} bytes Found: {}".format(
                    self.getSize(), len(data[offset:])
                )
            )
        for key, val in self.enum_dict().items():
            if int_val == val:
                self.val = key
                break
        # Value not found, invalid enumeration value
        else:
            raise TypeRangeException(int_val)

    def getSize(self):
        """ Calculates the size based on the size of an integer used to store it """
        return struct.calcsize(">i")



"""
(test) fprime.fbuild.settings:

Tests the F prime settings module.
@author joshuaa
"""

import pytest
import os
from pathlib import Path

from fprime.fbuild.settings import IniSettings

LOCAL_PATH = Path(__file__).parent


def full_path(path):
    path = LOCAL_PATH / Path(path)
    return path.resolve()


def test_settings():
    test_cases = [
        {
            "file": "nonexistent.ini",
            "expected": {
                "framework_path": full_path("../../../../.."),
                "install_dest": full_path("settings-data/build-artifacts"),
            },
        },
        {
            "file": "settings-empty.ini",
            "expected": {
                "settings_file": full_path("settings-data/settings-empty.ini"),
                "default_toolchain": "native",
                "default_ut_toolchain": "native",
                "framework_path": full_path("../../../../.."),
                "install_dest": full_path("settings-data/build-artifacts"),
                "library_locations": [],
                "environment_file": full_path("settings-data/settings-empty.ini"),
                "environment": {},
            },
        },
        {
            "file": "settings-custom-install.ini",
            "expected": {
                "settings_file": full_path("settings-data/settings-custom-install.ini"),
                "default_toolchain": "native",
                "default_ut_toolchain": "native",
                "framework_path": full_path("../../../../.."),
                "install_dest": full_path("test"),
                "library_locations": [],
                "environment_file": full_path(
                    "settings-data/settings-custom-install.ini"
                ),
                "environment": {},
            },
        },
        {
            "file": "settings-custom-toolchain.ini",
            "expected": {
                "settings_file": full_path("settings-data/settings-custom-toolchain.ini"),
                "default_toolchain": "custom1",
                "default_ut_toolchain": "custom2",
                "framework_path": full_path("../../../../.."),
                "install_dest": full_path("settings-data/build-artifacts"),
                "library_locations": [],
                "environment_file": full_path(
                    "settings-data/settings-custom-toolchain.ini"
                ),
                "environment": {},
            },
        },
    ]

    for case in test_cases:
        fp = full_path("settings-data/" + case["file"])
        print(fp)
        results = IniSettings.load(fp, LOCAL_PATH)
        assert case["expected"] == results, "{}: Expected {}, got {}".format(
            fp, case["expected"], results
        )



""" Cookie cutter wrapper used to template out components
"""
import os
import glob
import sys
from pathlib import Path
from typing import Dict

from cookiecutter.main import cookiecutter
from cookiecutter.exceptions import OutputDirExistsException

from fprime.fbuild.builder import Build, Target
from fprime.fbuild.cmake import CMakeExecutionException


def confirm(msg):
    """ Confirms the removal of the file with a yes or no input """
    # Loop "forever" intended
    while True:
        confirm_input = input(msg)
        if confirm_input.lower() in ["y", "yes"]:
            return True
        if confirm_input.lower() in ["n", "no"]:
            return False
        print("{} is invalid.  Please use 'yes' or 'no'".format(confirm_input))


def run_impl(deployment: Path, path: Path, platform: str, verbose: bool):
    """ Run implementation of files one time """
    target = Target.get_target("impl", set())
    build = Build(target.build_type, deployment, verbose=verbose)
    build.load(path, platform)

    hpp_files = glob.glob("{}/*.hpp".format(path), recursive=False)
    cpp_files = glob.glob("{}/*.cpp".format(path), recursive=False)
    cpp_files.sort(key=len)

    # Check destinations
    if not hpp_files or not cpp_files:
        print(
            "[WARNING] Failed to find .cpp and .hpp destination files for implementation."
        )
        return False

    common = [name for name in cpp_files if "Common" in name] + []
    hpp_dest = hpp_files[0]
    cpp_dest = common[0] if common else cpp_files[0]

    if not confirm(
        "Generate implementations and merge into {} and {}?".format(hpp_dest, cpp_dest)
    ):
        return False
    build.execute(target, context=path, make_args={})

    hpp_files_template = glob.glob("{}/*.hpp-template".format(path), recursive=False)
    cpp_files_template = glob.glob("{}/*.cpp-template".format(path), recursive=False)

    if not hpp_files_template or not cpp_files_template:
        print("[WARNING] Failed to find generated .cpp-template or .hpp-template files")
        return False

    hpp_src = hpp_files_template[0]
    cpp_src = cpp_files_template[0]

    # Copy files without headers
    for src, dest in [(hpp_src, hpp_dest), (cpp_src, cpp_dest)]:
        with open(src, "r") as file_handle:
            lines = file_handle.readlines()
        lines = lines[11:]  # Remove old header
        with open(dest, "a") as file_handle:
            file_handle.write("".join(lines))
    removals = [path / "mod.mk", path / "Makefile", Path(hpp_src), Path(cpp_src)]
    for removal in removals:
        os.remove(removal)
    return True


def add_to_cmake(list_file: Path, comp_path: Path):
    """ Adds new component to CMakeLists.txt"""
    print("[INFO] Found CMakeLists.txt at '{}'".format(list_file))
    with open(list_file, "r") as file_handle:
        lines = file_handle.readlines()
    topology_lines = [
        (line, text) for line, text in enumerate(lines) if "/Top/" in text
    ]
    line = len(topology_lines)
    if topology_lines:
        line, text = topology_lines[0]
        print(
            "[INFO] Topology inclusion '{}' found on line {}.".format(
                text.strip(), line + 1
            )
        )
    if not confirm(
        "Add component {} to {} {}?".format(
            comp_path,
            list_file,
            "at end of file" if not topology_lines else " before topology inclusion",
        )
    ):
        return False

    addition = 'add_fprime_subdirectory("${{CMAKE_CURRENT_LIST_DIR}}/{}/")\n'.format(
        comp_path
    )
    lines.insert(line, addition)
    with open(list_file, "w") as file_handle:
        file_handle.write("".join(lines))
    return True


def find_nearest_cmake_lists(component_dir: Path, deployment: Path, proj_root: Path):
    """Find the nearest CMakeLists.txt file

    The "nearest" file is defined as the closes parent that is not the "project root" that contains a CMakeLists.txt. If
    none is found, the same procedure is run from the deployment directory and includes the project root this time. If
    nothing is found, None is returned.

    In short the following in order of preference:
     - Any Component Parent
     - Any Deployment Parent
     - Project Root
     - None

    Args:
        component_dir: directory of new component
        deployment: deployment directory
        proj_root: project root directory

    Returns:
        path to CMakeLists.txt or None
    """
    test_path = component_dir.parent
    # First iterate from where we are, then from the deployment to find the nearest CMakeList.txt nearby
    for test_path, end_path in [(test_path, proj_root), (deployment, proj_root.parent)]:
        while proj_root is not None and test_path != proj_root.parent:
            cmake_list_file = test_path / "CMakeLists.txt"
            if cmake_list_file.is_file():
                return cmake_list_file
            test_path = test_path.parent
    return None


def new_component(
    path: Path, deployment: Path, platform: str, verbose: bool, settings: Dict[str, str]
):
    """ Uses cookiecutter for making new components """
    try:
        print("[WARNING] **** fprime-util new is prototype functionality ****")
        calculated_defaults = {}
        proj_root = None
        try:
            proj_root = Path(settings.get("project_root", None))
            print(proj_root)
            comp_parent_path = path.relative_to(proj_root)
            back_path = os.sep.join([".." for _ in str(comp_parent_path).split(os.sep)])
            calculated_defaults["component_path"] = str(comp_parent_path).rstrip(os.sep)
            calculated_defaults["component_path_to_fprime_root"] = str(
                back_path
            ).rstrip(os.sep)
        except (ValueError, TypeError):
            print(
                "[WARNING] No found project root. Set 'component_path' and 'component_path_to_fprime_root' carefully"
            )
        source = "gh:SterlingPeet/cookiecutter-fprime-component"
        print("[INFO] Cookiecutter source: {}".format(source))
        print()
        print("----------------")
        print(
            "[INFO] Help available here: https://github.com/SterlingPeet/cookiecutter-fprime-component/blob/master/README.rst#id3"
        )
        print("----------------")
        print()
        final_component_dir = Path(
            cookiecutter(source, extra_context=calculated_defaults)
        ).resolve()
        if proj_root is None:
            print(
                "[INFO] Created component directory without adding to build system nor generating implementation {}".format(
                    final_component_dir
                )
            )
            return 0
        # Attempt to register to CMakeLists.txt
        cmake_lists_file = find_nearest_cmake_lists(
            final_component_dir, deployment, proj_root
        )
        if cmake_lists_file is None or not add_to_cmake(
            cmake_lists_file, final_component_dir.relative_to(cmake_lists_file.parent)
        ):
            print(
                "[INFO] Could not register {} with build system. Please add it and generate implementations manually.".format(
                    final_component_dir
                )
            )
            return 0
        # Attempt implementation
        if not run_impl(deployment, final_component_dir, platform, verbose):
            print(
                "[INFO] Could not generate implementations. Please do so manually.".format(
                    final_component_dir
                )
            )
            return 0
        print("[INFO] Created new component and created initial implementations.")
        print(
            "[INFO] Next run `fprime-util build{}` in {}".format(
                "" if platform is None else " " + platform, final_component_dir
            )
        )
        return 0
    except OutputDirExistsException as out_directory_error:
        print("{}".format(out_directory_error), file=sys.stderr)
    except CMakeExecutionException as exc:
        print("[ERROR] Failed to create component. {}".format(exc), file=sys.stderr)
    except OSError as ose:
        print("[ERROR] {}".format(ose))
    return 1



"""
fprime.util.build_helper.py

This script is defined to help users run standard make processes using CMake. This will support migrants from the old
make system, as well as enable more efficient practices when developing in the CMake system. The following functions
are supported herein:

 - build: build the current directory/module/deployment
 - impl: make implementation templates
 - testimpl: make testing templates
 - build_ut: build the current UTs
 - check: run modules unit tests

@author mstarch
"""
import argparse
import os
import re
import sys
from pathlib import Path
from typing import Dict, List, Tuple

from fprime.common.error import FprimeException
from fprime.fbuild.builder import (
    Target,
    Build,
    BuildType,
    GenerateException,
    InvalidBuildCacheException,
    UnableToDetectDeploymentException,
)
from fprime.fbuild.interaction import confirm, new_component
from fprime.fbuild.settings import IniSettings

CMAKE_REG = re.compile(r"-D([a-zA-Z0-9_]+)=(.*)")


def get_target(parsed: argparse.Namespace) -> Target:
    """Gets the target given the argparse namespace

    Takes the parsed namespace and processes it to a known matching target.

    Args:
        parsed: argparse namespace to read values from

    Returns:
        target that can support the given supplied namespace
    """
    mnemonic = parsed.command
    flags = {
        flag for flag in Target.get_all_possible_flags() if getattr(parsed, flag, False)
    }
    return Target.get_target(mnemonic, flags)


def validate(parsed, unknown):
    """
    Validate rules to ensure that the args are properly consistent. This will also generate a set of validated arguments
    to pass to CMake. This allows these values to be created, defaulted, and validated in one place
    :param parsed: args to validate
    :param unknown: unknown arguments
    :return: cmake arguments to pass to CMake
    """
    cmake_args = {}
    make_args = {}
    # Check platforms for existing toolchain, unless the default is specified.
    if parsed.command == "generate":
        d_args = {
            match.group(1): match.group(2)
            for match in [CMAKE_REG.match(arg) for arg in unknown]
        }
        cmake_args.update(d_args)
    # Build type only for generate, jobs only for non-generate
    elif parsed.command not in ["info", "purge", "hash-to-file", "new"]:
        parsed.settings = None  # Force to load from cache if possible
        make_args.update({"--jobs": (1 if parsed.jobs <= 0 else parsed.jobs)})
    return cmake_args, make_args


def add_target_parser(
    target: Target,
    subparsers,
    common: argparse.ArgumentParser,
    existing: Dict[str, Tuple[argparse.ArgumentParser, List[str]]],
):
    """Add a subparser for a given build target

    For a given build target, construct an argument parser with mnemonic as subcommand and needed flags. Then add it as
    a subparser to the global parent. If it already exists in existing, then just add non-existent flags/

    Args:
        target:     target for building a new subparser
        subparsers: argument parser to add a subparser to
        common:     common subparser to be used as parent carrying common flags
        existing:   dictionary storing the mnemonic to parser and flags tuple

    Notes:
        This functions has side effects of editing existing and the list of subparsers
    """
    if target.mnemonic not in existing:
        parser = subparsers.add_parser(
            target.mnemonic,
            parents=[common],
            add_help=False,
            help="{} in the specified directory".format(target.desc),
        )
        # --ut flag also exists at the global parsers, skip adding it
        existing[target.mnemonic] = (parser, ["ut"])
        # Common target-only items
        parser.add_argument(
            "-j",
            "--jobs",
            default=1,
            type=int,
            help="Parallel build job count. Default: %(default)s.",
        )
    parser, flags = existing[target.mnemonic]
    new_flags = [flag for flag in target.flags if flag not in flags]
    for flag in new_flags:
        parser.add_argument(
            "--{}".format(flag), action="store_true", default=False, help=target.desc
        )
    flags.extend(new_flags)


def parse_args(args):
    """
    Parse the arguments to the CLI. This will then enable the user to run the above listed commands via the commands.
    :param args: CLI arguments to process
    :return: parsed arguments in a Namespace
    """
    # Common parser specifying common arguments input into the utility
    common_parser = argparse.ArgumentParser(
        description="Common Parser for Common Ingredients."
    )
    common_parser.add_argument(
        "platform",
        nargs="?",
        default="default",
        help="F prime build platform (e.g. Linux, Darwin). Default specified in settings.ini",
    )
    common_parser.add_argument(
        "-d",
        "--deploy",
        dest="deploy",
        default=None,
        help="F prime deployment directory to use. May contain multiple build directories.",
    )
    common_parser.add_argument(
        "-p",
        "--path",
        default=os.getcwd(),
        help="F prime directory to operate on. Default: cwd, %(default)s.",
    )
    common_parser.add_argument(
        "-v",
        "--verbose",
        default=False,
        action="store_true",
        help="Turn on verbose output.",
    )
    common_parser.add_argument(
        "--ut",
        action="store_true",
        help="Run command against unit testing build type",
    )

    # Main parser for the whole application
    parsers = {}
    parser = argparse.ArgumentParser(description="F prime helper application.")
    subparsers = parser.add_subparsers(
        description="F prime utility command line. Please run one of the commands. "
        + "For help, run a command with the --help flag.",
        dest="command",
    )
    # Add non-target parsers
    generate_parser = subparsers.add_parser(
        "generate",
        help="Generate a build cache directory. Defaults to generating a release build cache",
        parents=[common_parser],
        add_help=False,
    )
    generate_parser.add_argument(
        "-Dxyz",
        action="append",
        help="Pass -D flags through to CMakes",
        nargs=1,
        default=[],
    )
    purge_parser = subparsers.add_parser(
        "purge",
        help="Purge build cache directories",
        add_help=False,
        parents=[common_parser],
    )
    purge_parser.add_argument(
        "-f",
        "--force",
        default=False,
        action="store_true",
        help="Purges the build directory by force. No confirmation will be requested.",
    )
    # Add a search for hash function
    hash_parser = subparsers.add_parser(
        "hash-to-file",
        help="Converts F prime build hash to filename.",
        parents=[common_parser],
        add_help=False,
    )
    hash_parser.add_argument(
        "hash",
        type=lambda x: int(x, 0),
        help="F prime assert hash to associate with a file.",
    )

    # Add a search for hash function
    subparsers.add_parser(
        "info",
        help="Gets fprime-util contextual information.",
        parents=[common_parser],
        add_help=False,
    )
    # New functionality
    subparsers.add_parser(
        "new",
        help="Generate a new component",
        parents=[common_parser],
        add_help=False,
    )
    for target in Target.get_all_targets():
        add_target_parser(target, subparsers, common_parser, parsers)
    # Parse and prepare to run
    parsed, unknown = parser.parse_known_args(args)
    bad = [bad for bad in unknown if not CMAKE_REG.match(bad)]
    if not hasattr(parsed, "command") or parsed.command is None:
        parser.print_help()
        sys.exit(1)
    elif bad:
        print("[ERROR] Unknown arguments: {}".format(", ".join(bad)))
        parser.print_help()
        sys.exit(1)
    cmake_args, make_args = validate(parsed, unknown)
    return parsed, cmake_args, make_args, parser


def print_info(parsed, deployment):
    """ Builds and prints the informational output block """
    cwd = Path(parsed.path)
    build_types = BuildType

    # Roll up targets for more concise display
    build_infos = {}
    local_generic_targets = set()
    global_generic_targets = set()
    # Loop through available builds and harvest targets
    for build_type in build_types:
        build = Build(build_type, deployment, verbose=parsed.verbose)
        try:
            build.load(cwd, parsed.platform)
        except InvalidBuildCacheException:
            print(
                "[WARNING] Not displaying results for build type '{}', missing build cache.".format(
                    build_type.get_cmake_build_type()
                )
            )
            continue
        build_info = build.get_build_info(cwd)
        # Target list
        local_targets = {
            "'{}'".format(target) for target in build_info.get("local_targets", [])
        }
        global_targets = {
            "'{}'".format(target) for target in build_info.get("global_targets", [])
        }
        build_artifacts = (
            build_info.get("auto_location")
            if build_info.get("auto_location") is not None
            else "N/A"
        )
        local_generic_targets = local_generic_targets.union(local_targets)
        global_generic_targets = global_generic_targets.union(global_targets)
        build_infos[build_type] = build_artifacts

    # Print out directory and deployment target sections
    print("[INFO] Fprime build information:")
    print("    Available directory targets: {}".format(" ".join(local_generic_targets)))
    print()
    print(
        "    Available deployment targets: {}".format(" ".join(global_generic_targets))
    )

    # Artifact locations come afterwards
    print("  ----------------------------------------------------------")
    for build_type, build_artifact_location in build_infos.items():
        format_string = "    {} build cache: {}"
        print(
            format_string.format(
                build_type.get_cmake_build_type(), build_artifact_location
            )
        )
    print()


def print_hash_info(lines, hash_val):
    """ Prints out hash info from lines """
    # Print out lines when found
    if lines:
        print("[INFO] File(s) associated with hash 0x{:x}".format(hash_val))
        for line in lines:
            print("   ", line, end="")
        return
    print(
        "[ERROR] No file hashes found in {} build. Do you need '--build-type Testing' for a unittest run?"
    )


def utility_entry(args):
    """ Main interface to F prime utility """
    parsed, cmake_args, make_args, parser = parse_args(args)
    cwd = Path(parsed.path)

    build_type = BuildType.BUILD_NORMAL
    if parsed.ut:
        build_type = BuildType.BUILD_TESTING

    try:
        deployment = (
            Path(parsed.deploy)
            if parsed.deploy is not None
            else Build.find_nearest_deployment(cwd)
        )
        if parsed.command is None:
            print("[ERROR] Must supply subcommand for fprime-util. See below.")
            parser.print_help()
            print_info(parsed, deployment)
        elif parsed.command == "info":
            print_info(parsed, deployment)
        elif parsed.command == "new":
            settings = IniSettings.load(deployment / "settings.ini", cwd)
            status = new_component(cwd, deployment, parsed.platform, parsed.verbose, settings)
            sys.exit(status)
        elif parsed.command == "hash-to-file":
            build = Build(build_type, deployment, verbose=parsed.verbose)
            lines = build.find_hashed_file(parsed.hash)
            print_hash_info(lines, parsed.hash)
        elif parsed.command == "generate":
            build = Build(build_type, deployment, verbose=parsed.verbose)
            build.invent(cwd, parsed.platform)
            toolchain = build.find_toolchain()
            print("[INFO] Generating build directory at: {}".format(build.build_dir))
            print(
                "[INFO] Using toolchain file {} for platform {}".format(
                    toolchain, parsed.platform
                )
            )
            if toolchain is not None:
                cmake_args.update({"CMAKE_TOOLCHAIN_FILE": toolchain})
            build.generate(cmake_args)
        elif parsed.command == "purge":
            for build_type in BuildType:
                build = Build(build_type, deployment, verbose=parsed.verbose)
                try:
                    build.load(cwd, parsed.platform)
                except InvalidBuildCacheException:
                    continue
                print(
                    "[INFO] {} build directory at: {}".format(
                        parsed.command.title(), build.build_dir
                    )
                )
                if parsed.force or confirm("Purge this directory (yes/no)?"):
                    build.purge()

            build = Build(BuildType.BUILD_NORMAL, deployment, verbose=parsed.verbose)
            try:
                build.load(cwd, parsed.platform)
            except InvalidBuildCacheException:
                # Just load the install destination regardless of whether the build is valid.
                pass

            install_dir = build.install_dest_exists()
            if install_dir is None:
                return

            print(
                "[INFO] {} install directory at: {}".format(
                    parsed.command.title(), install_dir
                )
            )
            if parsed.force or confirm("Purge installation directory (yes/no)?"):
                build.purge_install()
        else:
            target = get_target(parsed)
            build = Build(target.build_type, deployment, verbose=parsed.verbose)
            build.load(cwd, parsed.platform)
            build.execute(target, context=Path(parsed.path), make_args=make_args)
    except GenerateException as genex:
        print(
            "[ERROR] {}. Partial build cache remains. Run purge to clean-up.".format(
                genex
            ),
            file=sys.stderr,
        )
    except UnableToDetectDeploymentException:
        print(
            "[ERROR] Could not detect deployment directory for: {}".format(parsed.path)
        )
    except FprimeException as exc:
        print("[ERROR] {}".format(exc), file=sys.stderr)
        return 1
    return 0



#!/usr/bin/env python
####
# fprime Python Package:
#
# The F prime python package represents the core data types required to use or develop F prime
# python code. This includes both the F prime GDS and Test API, as well as the Autocoding tools.
# To install this package, run the following commands:
#
# User Install / Upgrade:
# ```
# pip install --upgrade fprime-tools
# ```
#
# Developer and Dynamic Installation:
# ```
# pip install -e ./Fw/Python
# ```
###
import os
from setuptools import find_packages, setup

# Setup a python package using setup-tools. This is a newer (and more recommended) technology
# then distutils.
setup(
    ####
    # Package Description:
    #
    # Basic package information. Describes the package and the data contained inside. This
    # information should match the F prime description information.
    ####
    name="fprime-tools",
    use_scm_version={
        "root": os.path.join("..", ".."),
        "relative_to": __file__,
    },
    license="Apache 2.0 License",
    description="F Prime Flight Software core data types",
    long_description="""
This package contains the necessary core data types used by F prime. Users who seek to develop tools,
utilities, and other libraries used to interact with F prime framework code can use these data types
to interact with the data coming from the FSW.
    """,
    url="https://github.com/nasa/fprime",
    keywords=["fprime", "embedded", "nasa"],
    project_urls={"Issue Tracker": "https://github.com/nasa/fprime/issues"},
    # Author of Python package, not F prime.
    author="Michael Starch",
    author_email="Michael.D.Starch@jpl.nasa.gov",
    ####
    # Included Packages:
    #
    # Will search for and included all python packages under the "src" directory.  The root package
    # is set to 'src' to avoid package names of the form src.fprime.
    ####
    packages=find_packages("src"),
    package_dir={"": "src"},
    ####
    # Classifiers:
    #
    # Standard Python classifiers used to describe this package.
    ####
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    # Requires Python 3.6+
    python_requires=">=3.6",
    install_requires=[
        "lxml==4.6.3",
        "Markdown==3.3.4",
        "pexpect==4.8.0",
        "pytest==6.2.4",
        "Cheetah3==3.2.6.post2",
        "setuptools-scm==6.0.1",
        "cookiecutter==1.7.2",
    ],
    extras_require={
        "dev": [
            "black==21.5b1",
            "pylama==7.7.1",
            "pylint==2.8.2",
            "pre-commit==2.12.1",
            "sphinx",
            "sphinxcontrib.mermaid",
            "sphinx-rtd-theme",
            "sphinx-autoapi",
            "sphinx-autoapi",
            "recommonmark",
        ]
    },
    # Setup and test requirements, not needed by normal install
    setup_requires=["pytest-runner==5.3.0", "setuptools_scm==6.0.1"],
    tests_require=["pytest"],
    # Create a set of executable entry-points for running directly from the package
    entry_points={
        "console_scripts": ["fprime-util = fprime.util.__main__:main"],
        "gui_scripts": [],
    },
)



"""Python implementations of built-in fprime types

This package provides the modules necessary for the representation of fprime types (Fw/Types) package in a Python
context. These standard types are represented as Python classes representing each individual type.

Example:
    U32 is represented by fprime.common.modules.serialize.numerical_types.U32Type

Special autocoded types are represented by configurable classes: ArrayType, SerializableType, and EnumType. These
represent a generic version of these classes whose designed properties are set as dynamic properties on the python
class itself.

Example:
    Serializables are represented by fprime.common.modules.serialize.serializable_type.SerializableType
    SerializableType(typename="MySerializable") is used to specify the Ai serializable's name
"""



"""
Created on Dec 18, 2014

@author: tcanham

"""
import struct

from fprime.constants import DATA_ENCODING

from . import type_base
from .type_exceptions import (
    DeserializeException,
    NotInitializedException,
    StringSizeException,
    TypeMismatchException,
)


class StringType(type_base.ValueType):
    """
    String type representation for F prime. This is a value type that stores a half-word first for representing the
    length of this given string.
    """

    def __init__(self, val=None, max_string_len=None):
        """
        Constructor to build a string
        @param val: value form which to create a string. Default: None.
        @param max_string: maximum length of the string. Default: None, not used.
        """
        self.__max_string_len = max_string_len
        super().__init__(val)

    def validate(self, val):
        """ Validates that this is a string """
        if not isinstance(val, str):
            raise TypeMismatchException(str, type(val))
        elif self.__max_string_len is not None and len(val) > self.__max_string_len:
            raise StringSizeException(len(self.val), self.__max_string_len)

    def serialize(self):
        """
        Serializes the string in a binary format
        """
        # If val is never set then it is init exception...
        if self.val is None:
            raise NotInitializedException(type(self))
        # Check string size before serializing
        elif (
            self.__max_string_len is not None and len(self.val) > self.__max_string_len
        ):
            raise StringSizeException(len(self.val), self.__max_string_len)
        # Pack the string size first then return the encoded data
        buff = struct.pack(">H", len(self.val)) + self.val.encode(DATA_ENCODING)
        return buff

    def deserialize(self, data, offset):
        """
        Deserializes a string from the given data buffer.
        """
        try:
            val_size = struct.unpack_from(">H", data, offset)[0]
            # Deal with not enough data left in the buffer
            if len(data[offset + 2 :]) < val_size:
                raise DeserializeException(
                    "Not enough data to deserialize string data. Needed: {} Left: {}".format(
                        val_size, len(data[offset + 2 :])
                    )
                )
            # Deal with a string that is larger than max string
            elif self.__max_string_len is not None and val_size > self.__max_string_len:
                raise StringSizeException(val_size, self.__max_string_len)
            self.val = data[offset + 2 : offset + 2 + val_size].decode(DATA_ENCODING)
        except struct.error:
            raise DeserializeException("Not enough bytes to deserialize string length.")

    def getSize(self):
        """
        Get the size of this object
        """
        return struct.calcsize(">H") + len(self.val)



"""
numerical_types.py:

A file that contains the definitions for all the integer types provided as part of F prime.  F prime supports integers
that map to stdint.h integer sizes, that is, 8-bit, 16-bit, 32-bit, and 64-bit signed and unsigned integers.

@author mstarch
"""
import abc
import re
import struct

from .type_base import ValueType
from .type_exceptions import (
    DeserializeException,
    NotInitializedException,
    TypeMismatchException,
    TypeRangeException,
)

BITS_RE = re.compile(r"[IUF](\d\d?)")


class NumericalType(ValueType, abc.ABC):
    """ Numerical types that can be serialized using struct and are of some power of 2 byte width """

    @classmethod
    def get_bits(cls):
        """ Gets the integer bits of a given type """
        match = BITS_RE.match(cls.__name__)
        assert (
            match
        ), "Type {} does not follow format I#Type U#Type nor F#Type required of numerical types".format(
            cls
        )
        return int(match.group(1))

    @classmethod
    def getSize(cls):
        """ Gets the size of the integer based on the size specified in the class name """
        return int(cls.get_bits() / 8)

    @staticmethod
    @abc.abstractmethod
    def get_serialize_format():
        """ Gets the format serialization string such that the class can be serialized via struct """

    def serialize(self):
        """ Serializes this type using struct and the val property """
        if self.val is None:
            raise NotInitializedException(type(self))
        return struct.pack(self.get_serialize_format(), self.val)

    def deserialize(self, data, offset):
        """ Serializes this type using struct and the val property """
        try:
            self.val = struct.unpack_from(self.get_serialize_format(), data, offset)[0]
        except struct.error as err:
            raise DeserializeException(str(err))


class IntegerType(NumericalType, abc.ABC):
    """ Base class that represents all integer common functions """

    def validate(self, val):
        """ Validates the given integer. """
        if not isinstance(val, int):
            raise TypeMismatchException(int, type(val))
        max_val = 1 << (
            self.get_bits() - (1 if self.__class__.__name__.startswith("I") else 0)
        )
        min_val = -max_val if self.__class__.__name__.startswith("I") else 0
        # Compare to min and max
        if val < min_val or val >= max_val:
            raise TypeRangeException(val)


class FloatType(NumericalType, abc.ABC):
    """ Base class that represents all float common functions """

    def validate(self, val):
        """ Validates the given integer. """
        if not isinstance(val, float):
            raise TypeMismatchException(float, type(val))


class I8Type(IntegerType):
    """ Single byte integer type. Represents C chars """

    @staticmethod
    def get_serialize_format():
        """ Allows serialization using struct """
        return "b"


class I16Type(IntegerType):
    """ Double byte integer type. Represents C shorts """

    @staticmethod
    def get_serialize_format():
        """ Allows serialization using struct """
        return ">h"


class I32Type(IntegerType):
    """ Four byte integer type. Represents C int32_t, """

    @staticmethod
    def get_serialize_format():
        """ Allows serialization using struct """
        return ">i"


class I64Type(IntegerType):
    """ Eight byte integer type. Represents C int64_t, """

    @staticmethod
    def get_serialize_format():
        """ Allows serialization using struct """
        return ">q"


class U8Type(IntegerType):
    """ Single byte integer type. Represents C chars """

    @staticmethod
    def get_serialize_format():
        """ Allows serialization using struct """
        return "B"


class U16Type(IntegerType):
    """ Double byte integer type. Represents C shorts """

    @staticmethod
    def get_serialize_format():
        """ Allows serialization using struct """
        return ">H"


class U32Type(IntegerType):
    """ Four byte integer type. Represents C unt32_t, """

    @staticmethod
    def get_serialize_format():
        """ Allows serialization using struct """
        return ">I"


class U64Type(IntegerType):
    """ Eight byte integer type. Represents C unt64_t, """

    @staticmethod
    def get_serialize_format():
        """ Allows serialization using struct """
        return ">Q"


class F32Type(FloatType):
    """ Eight byte integer type. Represents C unt64_t, """

    @staticmethod
    def get_serialize_format():
        """ Allows serialization using struct """
        return ">f"


class F64Type(FloatType):
    """ Eight byte integer type. Represents C unt64_t, """

    @staticmethod
    def get_serialize_format():
        """ Allows serialization using struct """
        return ">d"
